#region PDFsharp - A .NET library for processing PDF
//
// Authors:
//   David Stephensen
//
// Copyright (c) 2005-2016 empira Software GmbH, Cologne Area (Germany)
//
// http://www.pdfsharp.com
// http://sourceforge.net/projects/pdfsharp
//
// Permission is hereby granted, free of charge, to any person obtaining a
// copy of this software and associated documentation files (the "Software"),
// to deal in the Software without restriction, including without limitation
// the rights to use, copy, modify, merge, publish, distribute, sublicense,
// and/or sell copies of the Software, and to permit persons to whom the
// Software is furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included
// in all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
// THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
// FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
// DEALINGS IN THE SOFTWARE.

// Code derived from:
/*
 * Copyright (c) 1988-1997 Sam Leffler
 * Copyright (c) 1991-1997 Silicon Graphics, Inc.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * Sam Leffler and Silicon Graphics may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of Sam Leffler and Silicon Graphics.
 *
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
 *
 * IN NO EVENT SHALL SAM LEFFLER OR SILICON GRAPHICS BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE
 * OF THIS SOFTWARE.
 */

/*
 * TIFF Library.
 * Rev 5.0 Lempel-Ziv & Welch Compression Support
 *
 * This code is derived from the compress program whose code is
 * derived from software contributed to Berkeley by James A. Woods,
 * derived from original work by Spencer Thomas and Joseph Orost.
 *
 * The original Berkeley copyright notice appears below in its entirety.
 */

/*
 * Copyright (c) 1985, 1986 The Regents of the University of California.
 * All rights reserved.
 *
 * This code is derived from software contributed to Berkeley by
 * James A. Woods, derived from original work by Spencer Thomas
 * and Joseph Orost.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by the University of California, Berkeley.  The name of the
 * University may not be used to endorse or promote products derived
 * from this software without specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */
#endregion

using System;
using System.IO;

namespace PdfSharp.Pdf.Filters
{
    /// <summary>
    /// Implements the LzwDecode filter.
    /// </summary>
    public class LzwDecode : Filter
    {
        // Reference: 3.3.3  LZWDecode and FlateDecode Filters / Page 71

        /// <summary>
        /// Throws a NotImplementedException because the obsolete LZW encoding is not supported by PDFsharp.
        /// </summary>
        public override byte[] Encode(byte[] data)
        {
            throw new NotImplementedException("PDFsharp does not support LZW encoding.");
        }

        // The TIFF spec specifies that encoded bit strings range from 9 to 12 bits.
        const int BITS_MIN = 9;     // start with 9 bits
        const int BITS_MAX = 12;    // max of 12 bit strings
                                    // predefined codes
        const int CODE_CLEAR = 256; // code to clear string table
        const int CODE_EOI = 257;   // end-of-information code
        const int CODE_FIRST = 258; // first free code entry

        class BitReader
        {
            byte[] data;
            int dataPos;

            int nbits;      // # of bits/code
            int nextdata;   // next bits of i/o
            int nextbits;   // # of valid bits in nextdata
            long bitsleft;  // available bits in raw data

            int nbitsmask;  // nbits 1 bits, right adjusted

            public BitReader(byte[] data)
            {
                this.data = data;
                dataPos = 0;

                nextbits = 0;
                nextdata = 0;
                bitsleft = data.Length << 3;
            }

            public void ResetNumberOfBits()
            {
                SetNumberOfBits(BITS_MIN);
            }

            public void IncreaseNumberOfBits()
            {
                SetNumberOfBits(nbits + 1);
            }

            void SetNumberOfBits(int newNumBits)
            {
                nbits = newNumBits;
                if (nbits > BITS_MAX)
                {   // should not happen
                    nbits = BITS_MAX;
                }
                nbitsmask = (1 << nbits) - 1;
            }

            public int Position { get { return dataPos; } }

            public int BitMask { get { return nbitsmask; } }

            public int NextCode()
            {
                if (bitsleft < (long)nbits)
                {
                    /* TIFFWarningExt(
                        "LZWDecode: Strip %d not terminated with EOI code", 
                        tif_curstrip); */
                    return CODE_EOI;
                }
                else
                {
                    nextdata = (nextdata << 8) | data[dataPos++];
                    nextbits += 8;
                    if (nextbits < nbits)
                    {
                        nextdata = (nextdata << 8) | data[dataPos++];
                        nextbits += 8;
                    }
                    int code = (nextdata >> (nextbits - nbits)) & nbitsmask;
                    nextbits -= nbits;
                    bitsleft -= (long)nbits;
                    return code;
                }
            }
        }

        class CodeTable
        {
            struct CodeEntry
            {
                public int nextEntry;   // number of next entry, or -1 if none
                public int length;      // string len, including this token
                public byte value;      // data value
                public byte firstchar;  // first token of string
            };

            CodeEntry[] codetab = null; // kept separate for small machines

            int freeEntry;  // next free entry
            int maxcode;    // max available entry

            // CodeTable needs to talk to BitReader.
            BitReader bits;

            public CodeTable(BitReader bits)
            {
                this.bits = bits;

                codetab = new CodeEntry[1 << BITS_MAX];
                // Pre-load the table.
                for (int code = 0; code < 256; ++code)
                {
                    codetab[code].value = (byte)code;
                    codetab[code].firstchar = (byte)code;
                    codetab[code].length = 1;
                    codetab[code].nextEntry = -1;
                }
                // Zero out the unused entries.
                ClearEntries(CODE_CLEAR, CODE_FIRST);
                Reset();
            }

            // Zero entries that are not yet filled in. We do this to guard against bogus input data
            // that causes us to index into undefined entries. If you can come up with a way
            // to safely bounds-check input codes while decoding then you can remove this operation.
            public void ClearEntries(int start, int end)
            {
                for (int i = start; i < end; ++i)
                {
                    codetab[i].value = 0;
                    codetab[i].firstchar = 0;
                    codetab[i].length = 0;
                    codetab[i].nextEntry = -1;
                }
            }

            public void Reset()
            {
                bits.ResetNumberOfBits();
                maxcode = bits.BitMask - 1;
                freeEntry = CODE_FIRST;
                ClearEntries(CODE_FIRST, codetab.Length);
            }

            public void AddEntry(int code, int oldcode)
            {
                if (freeEntry < 0 || freeEntry >= codetab.Length)
                {
                    throw new InvalidDataException(String.Format("Corrupted LZW table at offset {0}", bits.Position));
                }

                if (oldcode < 0 || oldcode >= codetab.Length)
                {
                    throw new InvalidDataException(String.Format("Corrupted LZW table at offset {0}", bits.Position));
                }
                codetab[freeEntry].nextEntry = oldcode;
                codetab[freeEntry].firstchar = codetab[oldcode].firstchar;
                codetab[freeEntry].length = codetab[oldcode].length + 1;
                codetab[freeEntry].value =
                    (code < freeEntry) ? codetab[code].firstchar : codetab[oldcode].firstchar;
                if (++freeEntry > maxcode)
                {
                    bits.IncreaseNumberOfBits();
                    maxcode = bits.BitMask - 1;
                }
            }

            public byte[] OutputCode(int code)
            {
                // Code maps to a string, copy string value to output (written in reverse).
                if (codetab[code].length == 0)
                {
                    throw new InvalidDataException(String.Format("LzwDecode: Wrong length of decoded string: data probably corrupted at offset {0}", bits.Position));
                }

                int len = codetab[code].length;
                var buf = new byte[len];
                do
                {
                    buf[--len] = codetab[code].value;
                    code = codetab[code].nextEntry;
                } while (code >= 0 && len > 0);

                if (code >= 0)
                {
                    throw new InvalidDataException(String.Format("LzwDecode: Bogus encoding, loop in the code table; offset {0}", bits.Position));
                }

                return buf;
            }
        }

        /// <summary>
        /// Decodes the specified data.
        /// </summary>
        public override byte[] Decode(byte[] data, FilterParms parms)
        {
            // Check for old bit-reversed codes.
            if (data.Length >= 2 && data[0] == 0 && (data[1] & 0x1) != 0)
            {
                throw new NotImplementedException("Old-style LZW codes not supported");
            }

            MemoryStream result = new MemoryStream();

            var bits = new BitReader(data);
            var codeTable = new CodeTable(bits);	// needs to talk to BitReader

            int oldcode = -1;   // previously recognized code

            while (true)
            {
                int code = bits.NextCode();
                if (code == CODE_EOI)
                {
                    break;
                }
                if (code == CODE_CLEAR)
                {
                    codeTable.Reset();
                    oldcode = -1;
                    continue;
                }

                if (oldcode >= 0)
                {
                    // Add the new entry to the code table.
                    codeTable.AddEntry(code, oldcode);
                }

                oldcode = code;

                if (code >= 256)
                {
                    byte[] buf = codeTable.OutputCode(code);
                    result.Write(buf, 0, buf.Length);
                }
                else
                {
                    result.WriteByte((byte)code);
                }
            }

            return result.ToArray();
        }
    }
}
